pro gain_cal,x_var=x_var,cal_type=cal_type,smoothing=smoothing, $
  frames_per_page=frames_per_page, tel_bsl=tel_bsl, refant=refant, $
  loose=loose, poly=poly, connect=connect, preavg=preavg, $
  no_unwrap=no_unwrap, non_point=non_point, apply_all=apply_all, $
  _extra=extra_keywords
;yes
;=Task:GAIN_CAL --- To excute gain calibration (SMA wrapper)
;#Type: calib 
;+Use:
;      GAIN_CAL determines the gain solutions and apply, if needed,
;      the solutions to the visibility data. In GAIN_CAL, the first
;      job is to specify the gain calibrators and set their fluxes.
;      This must be done by hand because the SMA has no calibrator
;      database. Both antenna-based and baseline-based solution can
;      be derived.
;         There is a flagging option to pick out individual points.
;      By doing this, it is necessary to re-run  gain_cal in order
;      to re-fit the gains to the data ignoring the recently flagged
;      points.
;@x_var:   
;      header variable for x-ccord:
;         'int'   : integrations --- the default 
;         'hours' : obvering time in hours
;         'el'    : elevation angle
;         'ha'    : hour angle
;@cal_type: 
;      gain solution to be derived. 
;        'amp'    : amplitude only 
;        'pha'    : phase only
;        'amp,pha': both amplitude and phase
;      For the telescope(antenna)-based calibration, 
;      cal_type='amp' implies cal_type='amp,pha'
;@smoothing: 
;      time interval over which data is smoothed to derive
;      gain curves. 
;        Defulat:
;             smoothing = 20 for x_var='int'
;          or smoothing = 2  for x_var='hours'
;@frames_per_page: 
;      number of frames per page
;@tel_bsl:  
;      'baseline' :baseline-based solution and
;      'telescope':antenna-based solution
;      'baseline' will be the default, if not set
;@refant: 
;      reference antenna needed in the ant-based
;      solution derivation, NO default
;@loose:  
;      loose restriction for ant/baseline number
;      matching, see gain.pro for details
;@poly:   
;      order of polynomial gain curve fitting;
;      integer 0 or higher. The default gain fit
;      will be boxcar smoothing if poly is not set.
;@connect:
;      deriving gain curves simply by connecting
;      data points. This is equivalent to NO smoothing.
;@preavg:
;      to pre-(vector)-average data points in consecutive
;      integrations in order to get better S/N for solving
;      (in particular antenna-based) gain curves.
;@apply_all:
;      to apply the solution to the full dataset, ignoring
;      the filter set-up. By detault, the gain solution is
;      only applied to filtered data. This option is useful
;      when one use partial baselines to obtain antenna-based
;      gain solutions but would like to apply the solution
;      to the full dataset.
;@no_unwrap:
;      by default phases will be unwrapped. Set this flag
;      to deactivate phase unwrapping.
;@non_point: 
;      deriving gain curves using planets/satellites as calibrator.
;&history:
;--------------------------------------------------------------------
;       cykuo 19dec03 adapting the form of the header
;---------------------------------------------------------------------

common global
common data_set
common plo

result=gain_ini(30.,use,all_souids,all_sources,all_amps, $
   numbs_3mm,numbs_1mm,fluxes_3mm,fluxes_1mm)

if (e.debug) then begin
  print,'Array  all_souids: ',all_souids
  print,'Array all_sources: ',all_sources
  print,'Array flag of use: ',use
  print,'Array  fluxes_3mm: ',fluxes_3mm
  print,'Array  fluxes_1mm: ',fluxes_1mm
endif

j = where(use eq 1)
ncals = n_elements(j)
print,'number of gain cals ',ncals
print,'gain cal source(s): ',all_sources[j]
print,'gain cal 3mm fluxes:',fluxes_3mm[j]
print,'gain cal 1mm fluxes:',fluxes_1mm[j]


gai_souids = make_array(ncals,/float)
gai_fluxes_3mm = make_array(ncals,/float)
gai_fluxes_1mm = make_array(ncals,/float)

for i = 0, ncals-1 do begin
  gai_souids[i] = all_souids[j[i]]
  gai_fluxes_3mm[i] = fluxes_3mm[j[i]]
  gai_fluxes_1mm[i] = fluxes_1mm[j[i]]
endfor

if (keyword_set(frames_per_page) eq 0) then frames_per_page = 1
if (keyword_set(x_var) eq 0) then x_var = 'int'
if (keyword_set(cal_type) eq 0) then cal_type = 'amp,pha'
if (n_elements(smoothing) eq 0) and (n_elements(poly) eq 0) then begin
  if x_var eq 'int' then smoothing = 20.
  if x_var eq 'hours' then smoothing = 2.
  if x_var eq 'el' then poly=2
endif
plid = 1


if not keyword_set(tel_bsl) then tel_bsl='baseline'
if (tel_bsl ne 'baseline' and tel_bsl ne 'telescope') then tel_bsl='baseline'
if (not keyword_set(connect)) then connect = 0

if (e.debug) then print,'----------CALLING GAIN----------'
if (keyword_set(connect)) then $
    result=gain(gai_souids,gai_fluxes_3mm,gai_fluxes_1mm, $
                tel_bsl,x_var,cal_type,frames_per_page, $
                dt_smooth=smoothing,plid,refant=refant,loose=loose, $
                connect=connect, preavg=preavg, no_unwrap=no_unwrap, $
                non_point=non_point,  _extra=extra_keywords) $
else $
  if (keyword_set(smoothing) gt 0) then $
    result=gain(gai_souids,gai_fluxes_3mm,gai_fluxes_1mm, $
                tel_bsl,x_var,cal_type,frames_per_page, $
                dt_smooth=smoothing,plid,refant=refant,loose=loose, $
                preavg=preavg, no_unwrap=no_unwrap, non_point=non_point, $
                 _extra=extra_keywords) $
  else $
    result=gain(gai_souids,gai_fluxes_3mm,gai_fluxes_1mm, $
                tel_bsl,x_var,cal_type,frames_per_page, $
                dt_smooth=-poly,plid, refant=refant,loose=loose, $
                preavg=preavg, no_unwrap=no_unwrap, non_point=non_point, $
                 _extra=extra_keywords)

if (result lt 0) then begin
   print,'**** No gain solution was derived ****'
endif else begin

  aa = ''
  read,aa,prompt='Apply gain solution? [NO <YES>]:  '
  if (aa eq 'YES' or aa eq 'yes' or aa eq 'Yes' or aa eq 'Y' or aa eq 'y') then begin
    print,'YES: apply gains'

    if (keyword_set(apply_all)) then begin
      print, ' --- Flag apply_all is set. Set filter temperarily to the full dataset ---'
      result=dat_filter(s_s,/save)
      result=dat_filter(s_f,/reset,/no_notify)
    endif

    print,'Applying Gains'
    result=cal_apply(gain=cal_type,x_var=x_var,tel_bsl=tel_bsl,refant=refant)

    if (keyword_set(apply_all)) then begin
      print,' --- Restoring filter ---'
      result=dat_filter(s_s,/restore)
    endif

  endif else begin
    print,'NO: nothing done'
  endelse

endelse

end
